@@ -1,8 +1,7 @@
-#!/usr/bin/env python
+#!/usr/bin/env python3
 # coding: utf8
-"""Python link-grammar test script"""
+"""Python3 link-grammar test script"""
 
-from __future__ import print_function
 import sys, os, re
 import locale
 import unittest
@@ -32,14 +31,14 @@ print(clg.linkgrammar_get_configuration())
 for imported_module in 'linkgrammar$', 'clinkgrammar', '_clinkgrammar', 'lg_testutils':
     module_found = False
     for module in sys.modules:
-        if re.search(r'^(linkgrammar\.)?'+imported_module, module):
+        if re.search(r'^(linkgrammar\.)?' + imported_module, module):
             print("Using", sys.modules[module], end='')
             if hasattr(sys.modules[module], '__version__'):
                 print(' version', sys.modules[module].__version__, end='')
             print()
             module_found = True
     if not module_found:
-        print("Warning: Module", imported_module,  "not loaded.")
+        print("Warning: Module", imported_module, "not loaded.")
 
 sys.stdout.flush()
 #===
@@ -373,10 +372,7 @@ class DBasicParsingTestCase(unittest.TestCase):
         self.assertTrue(isinstance(result[1], Linkage))
 
         # def test_unicode_encoded_string(self):
-        if is_python2():
-            result = self.parse_sent(u"I love going to the caf\N{LATIN SMALL LETTER E WITH ACUTE}.".encode('utf8'))
-        else:
-            result = self.parse_sent(u"I love going to the caf\N{LATIN SMALL LETTER E WITH ACUTE}.")
+        result = self.parse_sent(u"I love going to the caf\N{LATIN SMALL LETTER E WITH ACUTE}.")
         self.assertTrue(len(result) > 1)
         self.assertTrue(isinstance(result[0], Linkage))
         self.assertTrue(isinstance(result[1], Linkage))
@@ -414,9 +410,10 @@ class DBasicParsingTestCase(unittest.TestCase):
 
     def test_timer_exhausted_exception(self):
         self.assertRaises(LG_TimerExhausted,
-                          self.parse_sent,
-                          "This should take more than one second to parse! " * 20,
-                          ParseOptions(max_parse_time=1))
+                self.parse_sent,
+                "This sentence parses without null words, "
+                "and should take more than one second to parse!" * 14,
+                ParseOptions(max_parse_time=1,short_length=255,disjunct_cost=10.0,linkage_limit=10000))
 
 # The tests here are numbered since their order is important.
 # They depend on the result and state of the previous ones as follows:
@@ -425,7 +422,7 @@ class DBasicParsingTestCase(unittest.TestCase):
 class EErrorFacilityTestCase(unittest.TestCase):
     # Initialize class variables to invalid (for the test) values.
     handler = {
-        "default":  lambda x, y=None: None,
+        "default": lambda x, y=None: None,
         "previous": lambda x, y=None: None
     }
 
@@ -466,7 +463,7 @@ class EErrorFacilityTestCase(unittest.TestCase):
         self.assertEqual(self.__class__.handler["previous"].__name__, "error_handler_test")
         self.assertRaises(LG_Error, Dictionary, "seh_dummy2")
         self.gotit = None
-        for i in range(0, 2+self.testleaks):
+        for i in range(0, 2 + self.testleaks):
             self.numerr = LG_Error.printall(self.error_handler_test, self)
             if i == 0:
                 self.assertEqual(self.numerr, 1)
@@ -510,13 +507,13 @@ class EErrorFacilityTestCase(unittest.TestCase):
                                 self.errinfo, "bad param")
         self.assertRaisesRegexp(ValueError, "must be an integer",
                                 self.__class__.handler["default"],
-                                self.errinfo, clg.lg_None+1)
+                                self.errinfo, clg.lg_None + 1)
         self.assertRaises(ValueError, self.__class__.handler["default"],
                           self.errinfo, -1)
 
         try:
             self.param_ok = False
-            save_stdout  = divert_start(1) # Note: Handler parameter is stdout
+            save_stdout = divert_start(1) # Note: Handler parameter is stdout
             self.__class__.handler["default"](self.errinfo, 1)
             self.assertIn(dummy_lang, save_stdout.divert_end())
             self.param_ok = True
@@ -551,7 +548,7 @@ class EErrorFacilityTestCase(unittest.TestCase):
 
     def test_30_formatmsg(self):
         # Here the error handler is still set to None.
-        for _ in range (0, 1+self.testleaks):
+        for _ in range(0, 1 + self.testleaks):
             self.assertRaises(LG_Error, Dictionary, "formatmsg-test-dummy-dict")
             LG_Error.printall(self.error_handler_test, self)
             self.assertRegexpMatches(self.errinfo.formatmsg(), "link-grammar: Error: .*formatmsg-test-dummy-dict")
@@ -589,7 +586,7 @@ class EErrorFacilityTestCase(unittest.TestCase):
         # The error message is now visible (but we cannot test that).
         self.__class__.handler["previous"] = LG_Error.set_handler(self.__class__.handler["default"])
         self.assertIsNone(self.__class__.handler["previous"])
-        for _ in range(0, 1+self.testleaks):
+        for _ in range(0, 1 + self.testleaks):
             self.__class__.handler["previous"] = LG_Error.set_handler(self.__class__.handler["default"])
         self.assertEqual(self.__class__.handler["previous"].__name__, "_default_handler")
 
@@ -674,7 +671,7 @@ class HEnglishLinkageTestCase(unittest.TestCase):
     # -- word is/isn't split by suffix splitter
     # -- the one that is in the dict is not the grammatically appropriate word.
     #
-    # Let's is NOT split into two! Its in the dict as one word, lower-case only.
+    # Let's is NOT split into two! It's in the dict as one word, lower-case only.
     def test_f_captilization(self):
         self.assertEqual(list(self.parse_sent('Let\'s eat.')[0].words()),
              ['LEFT-WALL', 'let\'s', 'eat.v', '.', 'RIGHT-WALL'])
@@ -763,13 +760,13 @@ class HEnglishLinkageTestCase(unittest.TestCase):
         sent = 'Scientists sometimes may repeat experiments or use groups.'
         linkage = self.parse_sent(sent)[0]
         self.assertEqual(linkage.diagram(),
-"\n    +---------------------------------------Xp--------------------------------------+"
-"\n    +---------------------------->WV---------------------------->+                  |"
-"\n    |           +-----------------------Sp-----------------------+                  |"
-"\n    |           |                  +------------VJlpi------------+                  |"
-"\n    +---->Wd----+          +---E---+---I---+----Op----+          +VJrpi+---Op--+    |"
-"\n    |           |          |       |       |          |          |     |       |    |"
-"\nLEFT-WALL scientists.n sometimes may.v repeat.v experiments.n or.j-v use.v groups.n ."
+"\n    +----------------------------------------Xp---------------------------------------+"
+"\n    +---------------------------->WV---------------------------->+                    |"
+"\n    |                              +--------------I--------------+                    |"
+"\n    |           +--------Sp--------+       +<-------VJlpi<-------+                    |"
+"\n    +---->Wd----+          +---E---+       +----Op----+          +>VJrpi>+---Op--+    |"
+"\n    |           |          |       |       |          |          |       |       |    |"
+"\nLEFT-WALL scientists.n sometimes may.v repeat.v experiments.n or.j-v   use.v groups.n ."
 "\n\n")
         sent = 'I enjoy eating bass.'
         linkage = self.parse_sent(sent)[0]
@@ -781,7 +778,6 @@ class HEnglishLinkageTestCase(unittest.TestCase):
 "\nLEFT-WALL I.p enjoy.v eating.v bass.n-u ."
 "\n\n")
 
-
         sent = 'We are from the planet Gorpon'
         linkage = self.parse_sent(sent)[0]
         self.assertEqual(linkage.diagram(),
@@ -920,9 +916,6 @@ class ZENLangTestCase(unittest.TestCase):
 class JADictionaryLocaleTestCase(unittest.TestCase):
     @classmethod
     def setUpClass(cls):
-        if is_python2(): # Locale stuff seems to be broken
-            raise unittest.SkipTest("Test not supported with Python2")
-
         # python2: Gets system locale (getlocale() is not better)
         cls.oldlocale = locale.setlocale(locale.LC_CTYPE, None)
         #print('Current locale:', oldlocale)
@@ -955,9 +948,6 @@ class JADictionaryLocaleTestCase(unittest.TestCase):
 class JBDictCostReadingTestCase(unittest.TestCase):
     @classmethod
     def setUpClass(cls):
-        if is_python2(): # Locale stuff seems to be broken
-            raise unittest.SkipTest("Test not supported with Python2")
-
         cls.oldlocale = locale.setlocale(locale.LC_CTYPE, None)
         ru_locale = 'ru_RU.UTF-8' if os.name != 'nt' else 'Russian'
         try:
@@ -1096,7 +1086,13 @@ class ZRULangTestCase(unittest.TestCase):
              'облачк.=', '=а.ndnpi',
              '.', 'RIGHT-WALL'])
 
-def linkage_testfile(self, lgdict, popt, desc = ''):
+class ZXDictDialectTestCase(unittest.TestCase):
+    def test_dialect(self):
+        linkage_testfile(self, Dictionary(lang='en'), ParseOptions(dialect='headline'), 'dialect')
+
+#############################################################################
+
+def linkage_testfile(self, lgdict, popt, desc=''):
     """
     Reads sentences and their corresponding
     linkage diagrams / constituent printings.
@@ -1134,8 +1130,7 @@ def linkage_testfile(self, lgdict, popt, desc = ''):
 
     for line in parses:
         lineno += 1
-        if not is_python2():
-            line = line.decode('utf-8')
+        line = line.decode('utf-8')
 
         validate_opcode(ord(line[0])) # Use ord() for python2/3 compatibility
         if line[0] in 'INOCP':
@@ -1147,9 +1142,10 @@ def linkage_testfile(self, lgdict, popt, desc = ''):
             diagram = ""
             constituents = ""
             wordpos = ""
+            if popt.verbosity > 1:
+                print('Sentence:', sent)
             linkages = Sentence(sent, lgdict, popt).parse()
             linkage = next(linkages, None)
-            self.assertTrue(linkage, "at {}:{}: Sentence has no linkages".format(testfile, lineno))
 
         # Generate the next linkage of the last input sentence
         elif line[0] == 'N':
@@ -1163,9 +1159,14 @@ def linkage_testfile(self, lgdict, popt, desc = ''):
         # It ends with an empty line
         elif line[0] == 'O':
             diagram += line[1:]
-            if line[1] == '\n' and len(diagram) > 1:
-                self.assertEqual(linkage.diagram(), diagram, "at {}:{}".format(testfile, lineno))
-                diagram = None
+            if line[1] == '\n':
+                if diagram == 'C\nC\n':
+                    self.assertFalse(linkage)
+                    diagram = None
+                elif len(diagram) > 2:
+                    self.assertTrue(linkage, "at {}:{}: Sentence has no linkages".format(testfile, lineno))
+                    self.assertEqual(linkage.diagram(), diagram, "at {}:{}".format(testfile, lineno))
+                    diagram = None
 
         # Lines starting with C are the constituent output (type 1)
         # It ends with an empty line
@@ -1189,23 +1190,19 @@ def linkage_testfile(self, lgdict, popt, desc = ''):
 
         # Lines starting with "-" contain a Parse Option
         elif line[0] == '-':
-                exec('popt.' + line[1:]) in {}, locals()
+            exec('popt.' + line[1:]) in {}, locals()
 
         elif line[0] in '%\r\n':
             pass
         else:
             self.fail('\nTest file "{}": Invalid opcode "{}" (ord={})'.format(testfile, line[0], ord(line[0])))
 
-    self.assertIn(last_opcode , 'OCP', "Missing result comparison in " + testfile)
+    self.assertIn(last_opcode, 'OCP', "Missing result comparison in " + testfile)
 
 def warning(*msg):
     progname = os.path.basename(sys.argv[0])
     print("{}: Warning:".format(progname), *msg, file=sys.stderr)
 
-def is_python2():
-    return sys.version_info[:1] == (2,)
-
-
 import tempfile
 
 class divert_start(object):
@@ -1239,16 +1236,15 @@ class divert_start(object):
 lg_testutils.add_eqcost_linkage_order(Sentence)
 
 # For testing development branches, it may be sometimes useful to use the
-# "test", "debug" and "verbosity" options. The following allows specify them
+# "test", "debug" and "verbosity" options. The following allows to specify them
 # as "tests.py" arguments, interleaved with standard "unittest" arguments.
 
-for i,arg in enumerate(sys.argv):
-    # "-something=" confuses vim's syntax highlight, so add the '=' separately
+for i, arg in enumerate(sys.argv):
     debug = sys.argv.pop(i)[7:] if arg.startswith('-debug' + '=') else ''
-for i,arg in enumerate(sys.argv):
-    test = sys.argv.pop(i)[6:] if arg.startswith('-test' + '=')  else ''
-for i,arg in enumerate(sys.argv):
-    verbosity = int(sys.argv.pop(i)[11:]) if arg.startswith('-verbosity' + '=')  else ''
+for i, arg in enumerate(sys.argv):
+    test = sys.argv.pop(i)[6:] if arg.startswith('-test' + '=') else ''
+for i, arg in enumerate(sys.argv):
+    verbosity = int(sys.argv.pop(i)[11:]) if arg.startswith('-verbosity' + '=') else ''
 if (test or debug or verbosity):
     ParseOptions = lg_testutils.add_test_option(ParseOptions, test, debug, verbosity)
 
