@@ -14,21 +14,21 @@
 #include <string.h>
 
 #include "api-structures.h"
-#include "corpus/corpus.h"
+#include "dict-common/dict-defines.h"
+#include "dict-common/dialect.h"
 #include "dict-common/dict-utils.h"
-#include "disjunct-utils.h"   // for free_sentence_disjuncts()
+#include "disjunct-utils.h"             // free_sentence_disjuncts
 #include "linkage/linkage.h"
 #include "memory-pool.h"
-#include "parse/histogram.h"  // for PARSE_NUM_OVERFLOW
+#include "parse/histogram.h"            // PARSE_NUM_OVERFLOW
 #include "parse/parse.h"
-#include "post-process/post-process.h" // for post_process_new()
+#include "post-process/post-process.h"  // post_process_new
 #include "prepare/exprune.h"
 #include "string-set.h"
 #include "resources.h"
 #include "sat-solver/sat-encoder.h"
-#include "tokenize/spellcheck.h"
 #include "tokenize/tokenize.h"
-#include "tokenize/word-structures.h" // Needed for Word_struct
+#include "tokenize/word-structures.h"   // Word_struct
 #include "utilities.h"
 
 /* Its OK if this is racey across threads.  Any mild shuffling is enough. */
@@ -68,21 +68,6 @@ static int VDAL_compare_parse(Linkage l1, Linkage l2)
 	}
 }
 
-#ifdef USE_CORPUS
-static int CORP_compare_parse(Linkage l1, Linkage l2)
-{
-	Linkage_info * p1 = &l1->lifo;
-	Linkage_info * p2 = &l2->lifo;
-
-	double diff = p1->corpus_cost - p2->corpus_cost;
-
-	if (fabs(diff) < 1.0e-5)
-		return VDAL_compare_parse(p1, p2);
-	if (diff < 0.0) return -1;
-	return 1;
-}
-#endif
-
 /**
  * Create and initialize a Parse_Options object
  */
@@ -104,11 +89,8 @@ Parse_Options parse_options_create(void)
 
 	/* A cost of 2.7 allows the usual cost-2 connectors, plus the
 	 * assorted fractional costs, without going to cost 3.0, which
-	 * is used only during panic-parsing.
-	 * XXX In the long run, this should be fetched from the dictionary
-	 * (and should probably not be a parse option).
-	 */
-	po->disjunct_cost = 2.7;
+	 * is used only during panic-parsing. */
+	po->disjunct_cost = UNINITIALIZED_MAX_DISJUNCT_COST;
 	po->min_null_count = 0;
 	po->max_null_count = 0;
 	po->islands_ok = false;
@@ -120,32 +102,25 @@ Parse_Options parse_options_create(void)
 	po->use_spell_guess = 0;
 #endif /* defined HAVE_HUNSPELL || defined HAVE_ASPELL */
 
-#ifdef XXX_USE_CORPUS
-	/* Use the corpus cost model, if available.
-	 * It really does a better job at parse ranking.
-	 * Err .. sometimes ...
-	 */
-	po->cost_model.compare_fn = &CORP_compare_parse;
-	po->cost_model.type = CORPUS;
-#else /* USE_CORPUS */
 	po->cost_model.compare_fn = &VDAL_compare_parse;
 	po->cost_model.type = VDAL;
-#endif /* USE_CORPUS */
 	po->short_length = 16;
 	po->all_short = false;
 	po->perform_pp_prune = true;
 	po->twopass_length = 30;
 	po->repeatable_rand = true;
 	po->resources = resources_create();
-	po->use_cluster_disjuncts = false;
 	po->display_morphology = true;
+	po->dialect = (dialect_info){ .conf = strdup("") };
 
 	return po;
 }
 
+static void free_dialect_info(dialect_info *);
 int parse_options_delete(Parse_Options opts)
 {
 	resources_delete(opts->resources);
+	free_dialect_info(&opts->dialect);
 	free(opts);
 	return 0;
 }
@@ -157,16 +132,8 @@ void parse_options_set_cost_model_type(Parse_Options opts, Cost_Model_type cm)
 		opts->cost_model.type = VDAL;
 		opts->cost_model.compare_fn = &VDAL_compare_parse;
 		break;
-	case CORPUS:
-#ifdef USE_CORPUS
-		opts->cost_model.type = CORPUS;
-		opts->cost_model.compare_fn = &CORP_compare_parse;
-#else
-		prt_error("Error: Source code compiled with cost model 'CORPUS' disabled.\n");
-#endif
-		break;
 	default:
-		prt_error("Error: Illegal cost model: %d\n", cm);
+		prt_error("Error: Illegal cost model: %d\n", (int)cm);
 	}
 }
 
@@ -229,7 +196,17 @@ void parse_options_set_debug(Parse_Options opts, const char * dummy)
 }
 
 char * parse_options_get_debug(Parse_Options opts) {
-	return opts->debug;
+	static char buff[256]; /* see parse_options_set_debug() */
+	char *buffp = buff;
+
+	/* Remove the added commas. */
+	strcpy(buff, opts->debug);
+	if (buffp[0] == ',')
+		buffp++;
+	if ((buffp[0] != '\0') && buffp[strlen(buffp)-1] == ',')
+		buffp[strlen(buffp)-1] = '\0';
+
+	return buffp;
 }
 
 void parse_options_set_test(Parse_Options opts, const char * dummy)
@@ -266,7 +243,17 @@ void parse_options_set_test(Parse_Options opts, const char * dummy)
 }
 
 char * parse_options_get_test(Parse_Options opts) {
-	return opts->test;
+	static char buff[256]; /* see parse_options_set_test() */
+	char *buffp = buff;
+
+	/* Remove the added commas. */
+	strcpy(buff, opts->test);
+	if (buffp[0] == ',')
+		buffp++;
+	if ((buffp[0] != '\0') && buffp[strlen(buffp)-1] == ',')
+		buffp[strlen(buffp)-1] = '\0';
+
+	return buffp;
 }
 
 void parse_options_set_use_sat_parser(Parse_Options opts, bool dummy) {
@@ -388,14 +375,6 @@ int parse_options_get_max_memory(Parse_Options opts) {
 	return opts->resources->max_memory;
 }
 
-void parse_options_set_use_cluster_disjuncts(Parse_Options opts, bool dummy) {
-	opts->use_cluster_disjuncts = dummy;
-}
-
-bool parse_options_get_use_cluster_disjuncts(Parse_Options opts) {
-	return opts->use_cluster_disjuncts;
-}
-
 int parse_options_get_display_morphology(Parse_Options opts) {
 	return opts->display_morphology;
 }
@@ -420,6 +399,28 @@ void parse_options_reset_resources(Parse_Options opts) {
 	resources_reset(opts->resources);
 }
 
+/* Dialect. */
+
+static void free_dialect_info(dialect_info *dinfo)
+{
+	if (dinfo == NULL) return;
+
+	free(dinfo->cost_table);
+	dinfo->cost_table = NULL;
+	free((void *)dinfo->conf);
+}
+
+char * parse_options_get_dialect(Parse_Options opts)
+{
+	return opts->dialect.conf;
+}
+
+void parse_options_set_dialect(Parse_Options opts, const char *dconf)
+{
+	if (0 == strcmp(dconf, opts->dialect.conf)) return;
+	free_dialect_info(&opts->dialect);
+	opts->dialect.conf = strdup(dconf);
+}
 /***************************************************************
 *
 * Routines for creating destroying and processing Sentences
@@ -436,9 +437,6 @@ Sentence sentence_create(const char *input_string, Dictionary dict)
 	sent->dict = dict;
 	sent->string_set = string_set_create();
 	sent->rand_state = global_rand_state;
-	sent->E_list_pool = pool_new(__func__, "E_list", /*num_elements*/4096,
-	                             sizeof(E_list), /*zero_out*/false,
-	                             /*align*/false, /*exact*/false);
 	sent->Exp_pool = pool_new(__func__, "Exp", /*num_elements*/4096,
 	                             sizeof(Exp), /*zero_out*/false,
 	                             /*align*/false, /*exact*/false);
@@ -449,16 +447,25 @@ Sentence sentence_create(const char *input_string, Dictionary dict)
 	sent->postprocessor = post_process_new(dict->base_knowledge);
 
 	/* Make a copy of the input */
-	sent->orig_sentence = string_set_add (input_string, sent->string_set);
+	sent->orig_sentence = string_set_add(input_string, sent->string_set);
 
 	/* Set the minimum length for tracon sharing.
 	 * In that case a tracon list is produced for the pruning step,
 	 * and disjunct/connector packing is done too. */
-	sent->min_len_encoding = SENTENCE_MIN_LENGTH_TRAILING_HASH;
-	const char *min_len_encoding = test_enabled("len-trailing-hash");
+	if (IS_GENERATION(dict))
+		sent->min_len_encoding = 0;
+	else
+		sent->min_len_encoding = SENTENCE_MIN_LENGTH_TRAILING_HASH;
+	const char *min_len_encoding = test_enabled("min-len-encoding");
 	if (NULL != min_len_encoding)
 		sent->min_len_encoding = atoi(min_len_encoding+1);
 
+	/* Set the minimum length for pruning per null-count. */
+	sent->min_len_multi_pruning = SENTENCE_MIN_LENGTH_MULTI_PRUNING;
+	const char *min_len_multi_pruning = test_enabled("len-multi-pruning");
+	if (NULL != min_len_multi_pruning)
+		sent->min_len_multi_pruning = atoi(min_len_multi_pruning+1);
+
 	return sent;
 }
 
@@ -482,6 +489,9 @@ int sentence_split(Sentence sent, Parse_Options opts)
 		return -1;
 	}
 
+	if (!setup_dialect(dict, opts))
+		return -4;
+
 	/* Flatten the word graph created by separate_sentence() to a 2D-word-array
 	 * which is compatible to the current parsers.
 	 * This may fail if UNKNOWN_WORD is needed but
@@ -505,6 +515,8 @@ int sentence_split(Sentence sent, Parse_Options opts)
 		return -3;
 	}
 
+	if (verbosity >= D_USER_TIMES)
+		prt_error("#### Finished tokenizing (%zu tokens)\n", sent->length);
 	return 0;
 }
 
@@ -514,7 +526,7 @@ static void free_sentence_words(Sentence sent)
 	{
 		free(sent->word[i].alternatives);
 	}
-	free_sentence_disjuncts(sent);
+	free_sentence_disjuncts(sent, /*categories_too*/true);
 	free((void *) sent->word);
 	sent->word = NULL;
 }
@@ -529,15 +541,29 @@ void sentence_delete(Sentence sent)
 	free_linkages(sent);
 	post_process_free(sent->postprocessor);
 	post_process_free(sent->constituent_pp);
+	lg_exp_stringify(NULL);
+	free(sent->disjunct_used);
 
 	global_rand_state = sent->rand_state;
-	pool_delete(sent->fm_Match_node);
+	pool_delete(sent->Match_node_pool);
 	pool_delete(sent->Table_connector_pool);
-	pool_delete(sent->E_list_pool);
+	pool_delete(sent->wordvec_pool);
 	pool_delete(sent->Exp_pool);
 	pool_delete(sent->X_node_pool);
 	if (IS_DB_DICT(sent->dict))
+	{
+#if 0 /* Cannot reuse in case a previous sentence is not deleted yet. */
 		condesc_reuse(sent->dict);
+#endif
+		pool_reuse(sent->dict->Exp_pool);
+	}
+
+	if (NULL != sent->wildcard_word_dc_memblock)
+	{
+		free_categories_from_disjunct_array(sent->wildcard_word_dc_memblock,
+		                                    sent->wildcard_word_num_disjuncts);
+		free(sent->wildcard_word_dc_memblock);
+	}
 
 	free(sent);
 }
@@ -551,7 +577,7 @@ int sentence_length(Sentence sent)
 int sentence_null_count(Sentence sent)
 {
 	if (!sent) return 0;
-	return sent->null_count;
+	return (int)sent->null_count;
 }
 
 int sentence_num_linkages_found(Sentence sent)
@@ -603,6 +629,23 @@ int sentence_parse(Sentence sent, Parse_Options opts)
 {
 	int rc;
 
+	if (IS_GENERATION(sent->dict))
+	{
+		if (opts->use_sat_solver)
+		{
+			prt_error("Error: Cannot use the SAT parser in generation mode\n");
+			return -3;
+		}
+		if (opts->max_null_count > 0)
+		{
+			prt_error("Error: Cannot parse with nulls in generation mode\n");
+			return -3;
+		}
+	}
+
+	if (opts->disjunct_cost == UNINITIALIZED_MAX_DISJUNCT_COST)
+		opts->disjunct_cost = sent->dict->default_max_disjunct_cost;
+
 	sent->num_valid_linkages = 0;
 
 	/* If the sentence has not yet been split, do so now.
@@ -620,7 +663,7 @@ int sentence_parse(Sentence sent, Parse_Options opts)
 		 * garbage. Free it. We really should make the code that is panicking
 		 * do this free, but right now, they have no API for it, so we do it
 		 * as a favor. XXX FIXME someday. */
-		free_sentence_disjuncts(sent);
+		free_sentence_disjuncts(sent, /*categories_too*/true);
 	}
 
 	/* Check for bad sentence length */
@@ -638,6 +681,12 @@ int sentence_parse(Sentence sent, Parse_Options opts)
 	if (IS_DB_DICT(sent->dict))
 		condesc_setup(sent->dict);
 
+	for (WordIdx w = 0; w < sent->length; w++)
+	{
+		for (X_node *x = sent->word[w].x; x != NULL; x = x->next)
+			set_connector_farthest_word(x->exp, (int)w, (int)sent->length, opts);
+	}
+
 	/* Expressions were set up during the tokenize stage.
 	 * Prune them, and then parse.
 	 */
@@ -653,10 +702,10 @@ int sentence_parse(Sentence sent, Parse_Options opts)
 	}
 	print_time(opts, "Finished parse");
 
-	if ((verbosity > 0) &&
+	if ((verbosity > 0) && !IS_GENERATION(sent->dict) &&
 	   (PARSE_NUM_OVERFLOW < sent->num_linkages_found))
 	{
-		prt_error("Warning: Combinatorial explosion! nulls=%zu cnt=%d\n"
+		prt_error("Warning: Combinatorial explosion! nulls=%u cnt=%d\n"
 			"Consider retrying the parse with the max allowed disjunct cost set lower.\n"
 			"At the command line, use !cost-max\n",
 			sent->null_count, sent->num_linkages_found);
@@ -668,12 +717,30 @@ int sentence_parse(Sentence sent, Parse_Options opts)
  * Definitions for linkgrammar_get_configuration().
  */
 
+#define MACVAL(macro) #macro lg_str(=macro) " "
+
 #ifdef __STDC_VERSION__
-#define LG_S1 "__STDC_VERSION__=" lg_xstr(__STDC_VERSION__)
+#define LG_S1 MACVAL(__STDC_VERSION__)
 #else
 #define LG_S1
 #endif
 
+#ifdef _POSIX_C_SOURCE
+#define LG_S2 MACVAL(_POSIX_C_SOURCE)
+#else
+#define LG_S2
+#endif
+
+#if !defined _POSIX_C_SOURCE || _POSIX_C_SOURCE == 0
+ #ifdef _POSIX_SOURCE
+ #define LG_S3 MACVAL(_POSIX_SOURCE)
+ #else
+ #define LG_S3
+ #endif
+#else
+#define LG_S3
+#endif
+
 /* -DCC=$(CC) is added in the Makefile. */
 #ifdef CC
 #define LG_CC CC
@@ -684,19 +751,19 @@ int sentence_parse(Sentence sent, Parse_Options opts)
 #endif
 
 #ifdef __VERSION__
-#define LG_V1 "__VERSION__=" lg_xstr(__VERSION__)
+#define LG_V1 MACVAL(__VERSION__)
 #else
 #define LG_V1
 #endif
 
 #ifdef _MSC_FULL_VER
-#define LG_V2 "_MSC_FULL_VER=" lg_xstr(_MSC_FULL_VER)
+#define LG_V2 MACVAL(_MSC_FULL_VER)
 #else
 #define LG_V2
 #endif
 
 #define LG_COMP LG_CC " " LG_V1 " " LG_V2
-#define LG_STD LG_S1
+#define LG_STD LG_S1 LG_S2 LG_S3
 
 #ifdef __unix__
 #define LG_unix "__unix__ "
