@@ -17,23 +17,18 @@
 #include "api-types.h"                  // pp_knowledge
 #include "connectors.h"                 // ConTable
 #include "dict-structures.h"
+#include "memory-pool.h"                // Pool_desc
 #include "utilities.h"                  // locale_t
 
 #define EMPTY_CONNECTOR "ZZZ"
 #define UNLIMITED_CONNECTORS_WORD ("UNLIMITED-CONNECTORS")
 #define LIMITED_CONNECTORS_WORD ("LENGTH-LIMIT-")
+#define IS_GENERATION(dict) (dict->category != NULL)
 
 /* Forward decls */
 typedef struct Afdict_class_struct Afdict_class;
-typedef struct Exp_list_s Exp_list;
 typedef struct Regex_node_s Regex_node;
 
-/* Used for memory management */
-struct Exp_list_s
-{
-	Exp * exp_list;
-};
-
 typedef struct X_node_struct X_node;
 struct X_node_struct
 {
@@ -65,13 +60,30 @@ struct Afdict_class_struct
 };
 
 #define MAX_TOKEN_LENGTH 250     /* Maximum number of chars in a token */
-#define IDIOM_LINK_SZ 5
+#define IDIOM_LINK_SZ 16
 
-#ifdef HAVE_SQLITE
+#ifdef HAVE_SQLITE3
 #define IS_DB_DICT(dict) (NULL != dict->db_handle)
 #else
 #define IS_DB_DICT(dict) false
-#endif /* HAVE_SQLITE */
+#endif /* HAVE_SQLITE3 */
+
+/* "#define name value" */
+typedef struct
+{
+	String_id *set;                    /* "define" name set */
+	const char **name;
+	const char **value;
+	unsigned int size;                 /* Allocated value array size */
+} define_s;
+
+typedef struct
+{
+	String_id *set;                    /* Expression tag names */
+	const char **name;                 /* Tag name (indexed by tag id) */
+	unsigned int num;                  /* Number of tags */
+	unsigned int size;                 /* Allocated tag array size */
+} expression_tag;
 
 struct Dictionary_s
 {
@@ -81,8 +93,10 @@ struct Dictionary_s
 	const char * lang;
 	const char * version;
 	const char * locale;    /* Locale name */
+	double default_max_disjunct_cost;
 	locale_t     lctype;    /* Locale argument for the *_l() functions */
 	int          num_entries;
+	define_s     define;    /* Name-value definitions */
 
 	bool         use_unknown_word;
 	bool         unknown_word_defined;
@@ -90,6 +104,11 @@ struct Dictionary_s
 	bool         right_wall_defined;
 	bool         shuffle_linkages;
 
+	Dialect *dialect;                  /* "4.0.dialect" info */
+	expression_tag dialect_tag;        /* Expression dialect tag info */
+	expression_tag *macro_tag;         /* Macro tags for expression debug */
+	void *cached_dialect;              /* Only for dialect cache validation */
+
 	/* Affixes are used during the tokenization stage. */
 	Dictionary      affix_table;
 	Afdict_class *  afdict_class;
@@ -100,10 +119,7 @@ struct Dictionary_s
 
 	/* If not null, then use spelling guesser for unknown words */
 	void *          spell_checker;     /* spell checker handle */
-#if USE_CORPUS
-	Corpus *        corpus;            /* Statistics database */
-#endif
-#ifdef HAVE_SQLITE
+#ifdef HAVE_SQLITE3
 	void *          db_handle;         /* database handle */
 #endif
 
@@ -116,15 +132,17 @@ struct Dictionary_s
 
 	pp_knowledge  * base_knowledge;    /* Core post-processing rules */
 	pp_knowledge  * hpsg_knowledge;    /* Head-Phrase Structure rules */
-	Connector_set * unlimited_connector_set; /* NULL=everything is unlimited */
 	String_set *    string_set;        /* Set of link names in the dictionary */
 	Word_file *     word_file_header;
 	ConTable        contable;
 
-	/* exp_list links together all the Exp structs that are allocated
-	 * in reading this dictionary.  Needed for freeing the dictionary
-	 */
-	Exp_list        exp_list;
+	Pool_desc  * Exp_pool;
+
+	/* Sentence generation */
+	unsigned int num_categories;
+	unsigned int num_categories_alloced;
+	Category * category;      /* Word lists - indexed by category number */
+	bool generate_walls;      /* Generate walls too for wildcard words */
 
 	/* Private data elements that come in play only while the
 	 * dictionary is being read, and are not otherwise used.
@@ -139,13 +157,21 @@ struct Dictionary_s
 	char            current_idiom[IDIOM_LINK_SZ];
 	char            token[MAX_TOKEN_LENGTH];
 };
+
+bool is_stem(const char *);
+bool is_wall(const char *);
+bool is_macro(const char *);
+
+Exp *Exp_create(Pool_desc *);
+Exp *Exp_create_dup(Pool_desc *, Exp *);
+Exp *make_unary_node(Pool_desc *, Exp *);
+
+bool dictionary_generation_request(const Dictionary);
+
 /* The functions here are intended for use by the tokenizer, only,
  * and pretty much no one else. If you are not the tokenizer, you
  * probably don't need these. */
-
 bool dict_has_word(const Dictionary dict, const char *);
-Exp * Exp_create(Exp_list *);
 void add_empty_word(Sentence, X_node *);
-void free_Exp_list(Exp_list *);
 
 #endif /* _LG_DICT_COMMON_H_ */
