@@ -13,32 +13,12 @@
 #include "error.h"
 #include "dict-common/dict-common.h"
 #include "dict-common/dict-defines.h" // for MAX_WORD
+#include "dict-common/dict-utils.h" // for patch_subscript()
 #include "dict-common/file-utils.h"
 #include "string-set.h"
 #include "read-dict.h"
 #include "word-file.h"
 
-/** Replace the right-most dot with SUBSCRIPT_MARK */
-void patch_subscript(char * s)
-{
-	char *ds, *de;
-	int dp;
-	ds = strrchr(s, SUBSCRIPT_DOT);
-	if (!ds) return;
-
-	/* a dot at the end or a dot followed by a number is NOT
-	 * considered a subscript */
-	de = ds + 1;
-	if (*de == '\0') return;
-	dp = (int) *de;
-
-	/* If its followed by a UTF8 char, its NOT a subscript */
-	if (127 < dp || dp < 0) return;
-	/* assert ((0 < dp) && (dp <= 127), "Bad dictionary entry!"); */
-	if (isdigit(dp)) return;
-	*ds = SUBSCRIPT_MARK;
-}
-
 /**
  * Reads in one word from the file, allocates space for it,
  * and returns it.
@@ -94,9 +74,8 @@ Dict_node * read_word_file(Dictionary dict, Dict_node * dn, char * filename)
 		return NULL;
 	}
 
-	wf = (Word_file *) xalloc(sizeof (Word_file));
+	wf = malloc(sizeof (Word_file));
 	wf->file = string_set_add(filename, dict->string_set);
-	wf->changed = false;
 	wf->next = dict->word_file_header;
 	dict->word_file_header = wf;
 
@@ -107,7 +86,7 @@ Dict_node * read_word_file(Dictionary dict, Dict_node * dn, char * filename)
 			free_insert_list(dn);
 			return NULL;
 		}
-		Dict_node * dn_new = (Dict_node *) xalloc(sizeof(Dict_node));
+		Dict_node * dn_new = malloc(sizeof(Dict_node));
 		dn_new->left = dn;
 		dn = dn_new;
 		dn->string = s;
@@ -123,7 +102,7 @@ void free_Word_file(Word_file * wf)
 
 	for (;wf != NULL; wf = wf1) {
 		wf1 = wf->next;
-		xfree((char *) wf, sizeof(Word_file));
+		free(wf);
 	}
 }
 
