@@ -8,10 +8,12 @@
  * other process (machine-learning algo) to dynamically update
  * the dictionary.
  *
- * Copyright (c) 2014 Linas Vepstas <linasvepstas@gmail.com>
+ * Copyright (c) 2014,2021 Linas Vepstas <linasvepstas@gmail.com>
  */
 
-#ifdef HAVE_SQLITE
+#ifdef HAVE_SQLITE3
+
+#define D_SQL 5 /* Verbosity levels for this module are 5 and 6. */
 
 #include <sys/types.h>
 #include <sys/stat.h>
@@ -25,11 +27,12 @@
 #include "dict-common/dict-common.h"
 #include "dict-common/dict-impl.h"
 #include "dict-common/dict-structures.h"
-#include "dict-common/dict-utils.h"      // free_Exp()
+#include "dict-common/dict-utils.h"      // patch_subscript()
 #include "dict-common/file-utils.h"
 #include "dict-file/read-dict.h"         // dictionary_six()
+#include "error.h"
 #include "externs.h"
-#include "lg_assert.h"
+#include "memory-pool.h"
 #include "string-set.h"
 #include "tokenize/spellcheck.h"
 #include "utilities.h"
@@ -57,7 +60,7 @@ static const char * make_expression(Dictionary dict,
 	while (*p && (lg_isspace(*p))) p++;
 	if (0 == *p) return p;
 
-	/* If it's an open paren, assume its the begining of a new list */
+	/* If it's an open paren, assume its the beginning of a new list */
 	if ('(' == *p)
 	{
 		p = make_expression(dict, ++p, pex);
@@ -73,26 +76,25 @@ static const char * make_expression(Dictionary dict,
 				"Missing direction character in connector string: %s", con_start);
 
 		/* Create an expression to hold the connector */
-		Exp* e = malloc(sizeof(Exp));
+		Exp* e = Exp_create(dict->Exp_pool);
 		e->dir = *p;
 		e->type = CONNECTOR_type;
+		e->operand_next = NULL;
 		e->cost = 0.0;
-		e->u.l = NULL;
 		char * constr = NULL;
 		if ('@' == *con_start)
 		{
-			constr = strndup(con_start+1, p-con_start-1);
+			constr = strndupa(con_start+1, p-con_start-1);
 			e->multi = true;
 		}
 		else
 		{
-			constr = strndup(con_start, p-con_start);
+			constr = strndupa(con_start, p-con_start);
 			e->multi = false;
 		}
 
-		e->u.condesc = condesc_add(&dict->contable,
+		e->condesc = condesc_add(&dict->contable,
 		                           string_set_add(constr, dict->string_set));
-		free(constr);
 		*pex = e;
 	}
 
@@ -123,15 +125,15 @@ static const char * make_expression(Dictionary dict,
 	assert(NULL != rest, "Badly formed expression %s", exp_str);
 
 	/* Join it all together. */
-	Exp* join = malloc(sizeof(Exp));
+	Exp* join = Exp_create(dict->Exp_pool);
 	join->type = etype;
+	join->operand_next = NULL;
 	join->cost = 0.0;
-	E_list *ell = join->u.l = malloc(sizeof(E_list));
-	E_list *elr = ell->next = malloc(sizeof(E_list));
-	elr->next = NULL;
 
-	ell->e = *pex;
-	elr->e = rest;
+	join->operand_first = *pex;
+	(*pex)->operand_next = rest;
+	rest->operand_next = NULL;
+
 	*pex = join;
 
 	return p;
@@ -146,7 +148,9 @@ typedef struct
 	Dictionary dict;
 	Dict_node* dn;
 	bool found;
+	int count;
 	Exp* exp;
+	char* classname;
 } cbdata;
 
 static void db_free_llist(Dictionary dict, Dict_node *llist)
@@ -154,10 +158,7 @@ static void db_free_llist(Dictionary dict, Dict_node *llist)
 	Dict_node * dn;
 	while (llist != NULL)
 	{
-		Exp *e;
 		dn = llist->right;
-		e = llist->exp;
-		if (e) free_Exp(e);
 		free(llist);
 		llist = dn;
 	}
@@ -167,12 +168,13 @@ static void db_free_llist(Dictionary dict, Dict_node *llist)
 static int exp_cb(void *user_data, int argc, char **argv, char **colName)
 {
 	cbdata* bs = user_data;
+	Dictionary dict = bs->dict;
 
 	assert(2 == argc, "Bad column count");
 	assert(argv[0], "NULL column value");
 
 	Exp* exp = NULL;
-	make_expression(bs->dict, argv[0], &exp);
+	make_expression(dict, argv[0], &exp);
 	assert(NULL != exp, "Failed expression %s", argv[0]);
 
 	if (!strtodC(argv[1], &exp->cost))
@@ -182,23 +184,29 @@ static int exp_cb(void *user_data, int argc, char **argv, char **colName)
 		exp->cost = 1.0;
 	}
 
+	/* If the very first expression, just put it in place */
 	if (NULL == bs->exp)
 	{
 		bs->exp = exp;
 		return 0;
 	}
 
-	E_list *ell, *elr;
-	Exp* orn = malloc(sizeof(Exp));
-	orn->type = OR_type;
-	orn->cost = 0.0;
-	orn->u.l = ell = malloc(sizeof(E_list));
-	ell->next = elr = malloc(sizeof(E_list));
-	elr->next = NULL;
+	/* If the second expression, OR-it with the existing expression. */
+	if (OR_type != bs->exp->type)
+	{
+		Exp* orn = Exp_create(dict->Exp_pool);
+		orn->type = OR_type;
+		orn->cost = 0.0;
 
-	ell->e = bs->exp;
-	elr->e = exp;
-	bs->exp = orn;
+		orn->operand_first = exp;
+		exp->operand_next = bs->exp;
+		bs->exp = orn;
+		return 0;
+	}
+
+	/* Extend the OR-chain for the third and later expressions. */
+	exp->operand_next = bs->exp->operand_first;
+	bs->exp->operand_first = exp;
 
 	return 0;
 }
@@ -253,11 +261,8 @@ db_lookup_exp(Dictionary dict, const char *s, cbdata* bs)
 
 	if (es != s) free(es);
 
-	if (4 < verbosity)
-	{
-		printf("Found expression for class %s: ", s);
-		print_expression(bs->exp);
-	}
+	lgdebug(D_SQL+1, "Found expression for class %s: %s\n",
+	        s, lg_exp_stringify(bs->exp));
 }
 
 
@@ -346,12 +351,12 @@ static Dict_node * db_lookup_list(Dictionary dict, const char *s)
 	bs.dict = dict;
 	bs.dn = NULL;
 	db_lookup_common(dict, s, "=", morph_cb, &bs);
-	if (3 < verbosity)
+	if (verbosity_level(D_SQL))
 	{
 		if (bs.dn)
 		{
-			printf("Found expression for word %s: ", s);
-			print_expression(bs.dn->exp);
+			printf("Found expression for word %s: %s\n",
+	        s, lg_exp_stringify(bs.dn->exp));
 		}
 		else
 		{
@@ -373,12 +378,12 @@ static Dict_node * db_lookup_wild(Dictionary dict, const char *s)
 	bs.dict = dict;
 	bs.dn = NULL;
 	db_lookup_common(dict, s, "GLOB", morph_cb, &bs);
-	if (3 < verbosity)
+	if (verbosity_level(D_SQL))
 	{
 		if (bs.dn)
 		{
-			printf("Found expression for glob %s: ", s);
-			print_expression(bs.dn->exp);
+			printf("Found expression for glob %s: %s\n",
+			       s, lg_exp_stringify(bs.dn->exp));
 		}
 		else
 		{
@@ -388,6 +393,137 @@ static Dict_node * db_lookup_wild(Dictionary dict, const char *s)
 	return bs.dn;
 }
 
+/* ========================================================= */
+/* Callbacks and functions to support lexical category loading. */
+
+/* Used for `SELECT count(*) FROM foo` type of queries */
+static int count_cb(void *user_data, int argc, char **argv, char **colName)
+{
+	cbdata* bs = user_data;
+
+	assert(1 == argc, "Bad column count");
+	bs->count = atol(argv[0]);
+
+	return 0;
+}
+
+/* Record the name of each lexical class */
+static int classname_cb(void *user_data, int argc, char **argv, char **colName)
+{
+	cbdata* bs = user_data;
+	Dictionary dict = bs->dict;
+
+	/* Assuming here that a class name of a wall is the same as the wall
+	 * name, and a class name of a macro is in macro format. */
+	if (!dict->generate_walls && is_wall(argv[0])) return 0;
+	if (is_macro(argv[0])) return 0;
+
+	/* Add a category. */
+	/* This is intentionally off-by-one, per design. */
+	dict->num_categories++;
+	dict->category[dict->num_categories].num_words = 0;
+	dict->category[dict->num_categories].word = NULL;
+	dict->category[dict->num_categories].name =
+		string_set_add(argv[0], dict->string_set);
+
+	char category_string[16];     /* For the tokenizer - not used here */
+	snprintf(category_string, sizeof(category_string), " %x",
+	         dict->num_categories); /* ' ': See comment in build_disjuncts() */
+	string_set_add(category_string, dict->string_set);
+
+	return 0;
+}
+
+/* Record the words in each lexical class */
+static int classword_cb(void *user_data, int argc, char **argv, char **colName)
+{
+	cbdata* bs = user_data;
+	Dictionary dict = bs->dict;
+
+	char *word = strdupa(argv[0]);
+	patch_subscript(word);
+
+	/* Add the word. */
+	dict->category[dict->num_categories].word[bs->count] =
+		string_set_add(word, dict->string_set);
+	bs->count++;
+
+	return 0;
+}
+
+/* The current design for generation requires that all word categories
+ * be loaded into RAM before generation starts. This is required because
+ * a wild-card appearing in the generator forces a loop over all
+ * categories, so we may as well have them instantly available.
+ */
+static void add_categories(Dictionary dict)
+{
+	sqlite3 *db = dict->db_handle;
+	cbdata bs;
+	bs.dict = dict;
+
+	/* How many lexical categories are there? Find out. */
+	sqlite3_exec(db, "SELECT count(DISTINCT classname) FROM Disjuncts;",
+		count_cb, &bs, NULL);
+
+	dict->num_categories = 0;
+	dict->num_categories_alloced = 1 + bs.count + 1; // skip slot 0 + terminator
+	dict->category = malloc(dict->num_categories_alloced *
+	                        sizeof(*dict->category));
+
+	sqlite3_exec(db, "SELECT DISTINCT classname FROM Disjuncts;",
+		classname_cb, &bs, NULL);
+
+	/* Category 0 is unused, intentionally. Not sure why. */
+	unsigned int ncat = dict->num_categories;
+	for (unsigned int i=1; i<=ncat; i++)
+	{
+		/* For each category, get the expression. */
+		dyn_str *qry = dyn_str_new();
+		dyn_strcat(qry,
+			"SELECT disjunct, cost FROM Disjuncts WHERE classname = \'");
+		dyn_strcat(qry, dict->category[i].name);
+		dyn_strcat(qry, "\';");
+
+		bs.exp = NULL;
+		sqlite3_exec(db, qry->str, exp_cb, &bs, NULL);
+		dyn_str_delete(qry);
+
+		dict->category[i].exp = bs.exp;
+
+		/* ------------------ */
+		/* For each category, get the number of words in the category */
+		qry = dyn_str_new();
+		dyn_strcat(qry,
+			"SELECT count(*) FROM Morphemes WHERE classname = \'");
+		dyn_strcat(qry, dict->category[i].name);
+		dyn_strcat(qry, "\';");
+
+		sqlite3_exec(db, qry->str, count_cb, &bs, NULL);
+		dyn_str_delete(qry);
+
+		dict->category[i].num_words = bs.count;
+		dict->category[i].word =
+			malloc(bs.count * sizeof(dict->category[0].word));
+
+		/* ------------------ */
+		/* For each category, get the (subscripted) words in the category */
+		qry = dyn_str_new();
+		dyn_strcat(qry,
+			"SELECT subscript FROM Morphemes WHERE classname = \'");
+		dyn_strcat(qry, dict->category[i].name);
+		dyn_strcat(qry, "\';");
+
+		dict->num_categories = i;
+		bs.count = 0;
+		sqlite3_exec(db, qry->str, classword_cb, &bs, NULL);
+		dyn_str_delete(qry);
+	}
+
+	/* Set the termination entry. */
+	dict->category[dict->num_categories + 1].num_words = 0;
+}
+
 /* ========================================================= */
 /* Dictionary creation, setup, open procedures */
 
@@ -480,6 +616,10 @@ Dictionary dictionary_create_from_db(const char *lang)
 	dict->close = db_close;
 	condesc_init(dict, 1<<8);
 
+	dict->Exp_pool = pool_new(__func__, "Exp", /*num_elements*/4096,
+	                          sizeof(Exp), /*zero_out*/false,
+	                          /*align*/false, /*exact*/false);
+
 	/* Setup the affix table */
 	char *affix_name = join_path (lang, "4.0.affix");
 	dict->affix_table = dictionary_six(lang, affix_name, NULL, NULL, NULL, NULL);
@@ -493,9 +633,12 @@ Dictionary dictionary_create_from_db(const char *lang)
 	if (!afdict_init(dict))
 		goto failure;
 
-	dictionary_setup_locale(dict);
+	if (!dictionary_setup_defines(dict))
+		goto failure;
 
-	dictionary_setup_defines(dict);
+	/* Initialize word categories, for text generation. */
+	if (!dictionary_generation_request(dict))
+		add_categories(dict);
 
 	return dict;
 
@@ -504,4 +647,4 @@ failure:
 	return NULL;
 }
 
-#endif /* HAVE_SQLITE */
+#endif /* HAVE_SQLITE3 */
