@@ -23,43 +23,77 @@
 // Can undefine VERIFY_MATCH_LIST when done debugging...
 #define VERIFY_MATCH_LIST
 
+/* On a 64-bit machine, this struct should be exactly (6+2)*8=64 bytes long.
+ * Lets try to keep it that way (for performance). */
 struct Disjunct_struct
 {
+	/* 48 bytes of common stuff. */
 	Disjunct *next;
-	Disjunct *dup_table_next;
 	Connector *left, *right;
-	double cost;
-
-	/* match_left, right used only during parsing, for the match list. */
-	bool match_left, match_right;
-
+	gword_set *originating_gword; /* Set of originating gwords */
 	union
 	{
-		unsigned int match_id;     /* verify the match list integrity */
-		unsigned int dup_hash;     /* hash value for duplicate elimination */
+		struct
+		{
+			const char *word_string;  /* Subscripted dictionary word */
+			float cost;               /* Disjunct cost */
+			unsigned int is_category; /* Use the category field */
+		};
+		struct
+		{
+			/* Dictionary category & disjunct cost (for sentence generation). */
+			Category_cost *category;
+			unsigned int num_categories_alloced;
+			unsigned int num_categories;
+		};
 	};
-	gword_set *originating_gword; /* Set of originating gwords */
-	const char * word_string;     /* subscripted dictionary word */
+
+	/* Shared by different steps. For what | when. */
+	union
+	{
+		Disjunct *dup_table_next; /* Duplicate elimination | before pruning */
+		Disjunct *unused1;        /* Unused now | before parsing */
+	}; /* 8 bytes */
+
+	/* Shared by different steps. For what | when. */
+	union
+	{
+		uint32_t dup_hash;        /* Duplicate elimination | before pruning */
+		int32_t ordinal;          /* Generation mode | after d. elimination */
+	}; /* 4 bytes */
+
+	struct
+	{
+		bool match_left, match_right;
+#ifdef VERIFY_MATCH_LIST
+		uint16_t match_id;
+#endif
+	}; /* 4 bytes - during parsing */
 };
 
 /* Disjunct utilities ... */
 void free_disjuncts(Disjunct *);
-void free_sentence_disjuncts(Sentence);
+void free_sentence_disjuncts(Sentence, bool);
+void free_categories(Sentence);
+void free_categories_from_disjunct_array(Disjunct *, unsigned int);
 unsigned int count_disjuncts(Disjunct *);
 Disjunct * catenate_disjuncts(Disjunct *, Disjunct *);
-Disjunct * eliminate_duplicate_disjuncts(Disjunct *);
-char * print_one_disjunct(Disjunct *);
-void word_record_in_disjunct(const Gword *, Disjunct *);
+Disjunct * eliminate_duplicate_disjuncts(Disjunct *, bool);
 int left_connector_count(Disjunct *);
 int right_connector_count(Disjunct *);
 
 Tracon_sharing *pack_sentence_for_pruning(Sentence);
 Tracon_sharing *pack_sentence_for_parsing(Sentence);
 void free_tracon_sharing(Tracon_sharing *);
+void count_disjuncts_and_connectors(Sentence, unsigned int *, unsigned int *);
 
-void print_one_connector(Connector *, int, int);
-void print_connector_list(Connector *);
-void print_disjunct_list(Disjunct *);
+/* Print disjunct/connector */
+char *print_one_disjunct_str(const Disjunct *);
+char *print_one_connector_str(const Connector *, const char *);
+char *print_connector_list_str(const Connector *, const char *);
+void print_one_connector(const Connector *, const char *);
+void print_connector_list(const Connector *, const char *);
+void print_disjunct_list(const Disjunct *, const char *);
 void print_all_disjuncts(Sentence);
 
 /* Save and restore sentence disjuncts */
@@ -74,14 +108,8 @@ typedef struct
  * benefit of the pruning and parsing stages. To that end, unique tracons
  * are identified. For the parsing stage, a unique tracon_id is assigned
  * to the Connector's tracon_id field. For the pruning stage, this field
- * remains 0 (later used for pruning pass_count - see below) and instead
- * the Connector's refcount filed reflects the number of connectors which
- * are memory-shared.
- * In the pruning stage, on each pass the tracon_id of "good"
- * connectors is assigned the pass number so they will not be checked
- * again on the same pass. The pruning stage also uses the connector
- * refcount field as a reference count - the number of times this
- * connector is memory-shared.
+ * is unused and the Connector's refcount field reflects the number of
+ * connectors which are memory-shared.
  * The details of what considered a unique tracon is slightly different
  * for the pruning stage and the parsing stage - see the comment block
  * "Connector encoding, sharing and packing" in disjunct-utils.c.
@@ -97,7 +125,6 @@ typedef struct
 
 typedef struct
 {
-	unsigned int *num_cnctrs_per_word[2]; /* Indexed by word number */
 	/* Table of tracons. A 32bit index into the connector array in
 	 * memblock (instead of (Connector*)) is used for better use of the
 	 * CPU cache on 64-bit CPUs. */
@@ -110,25 +137,34 @@ struct tracon_sharing_s
 {
 	union
 	{
-		void *memblock;             /* Memory block for disjuncts & connectors */
-		Disjunct *dblock_base;      /* Start of disjunct block */
+		void *memblock;          /* Memory block for disjuncts & connectors */
+		Disjunct *dblock_base;   /* Start of disjunct block */
 	};
 	size_t memblock_sz;         /* memblock size */
 	Connector *cblock_base;     /* Start of connector block */
 	Connector *cblock;          /* Next available memory for connector */
 	Disjunct *dblock;           /* Next available memory for disjunct */
+	Disjunct **d;               /* The disjuncts (indexed by word number) */
 	unsigned int num_connectors;
 	unsigned int num_disjuncts;
 	Tracon_set *csid[2];        /* For generating unique tracon IDs */
 	int next_id[2];             /* Next unique tracon ID */
-	uintptr_t last_token;       /* Tracons are only unique per "token" */
+	uintptr_t last_token;       /* Tracons are the same only per this token */
 	int word_offset;            /* Start number for connector tracon_id */
-	bool is_pruning;            /* true - for pruning; false - for parsing */
+	bool is_pruning;            /* false: Parsing step, true: Pruning step */
 	Tracon_list *tracon_list;   /* Used only for pruning */
+
+	/* The number of different uppercase connector parts per side / word,
+	 * for sizing the prune power table. */
+	uint8_t *uc_seen[2];        /* The last word number in which an
+										  * uppercase connector part has been seen,
+										  * (hence doesn't need clearing between words).
+										  * Indexed by uc_num. */
+	unsigned int *num_cnctrs_per_word[2]; /* Indexed by word number */
 };
 
-void *save_disjuncts(Sentence, Tracon_sharing *, Disjunct **);
-void restore_disjuncts(Sentence, Disjunct **, void *, Tracon_sharing *);
+void *save_disjuncts(Sentence, Tracon_sharing *);
+void restore_disjuncts(Sentence, void *, Tracon_sharing *);
 void free_saved_disjuncts(Sentence);
 
 /** Get tracon by (dir, tracon_id). */
