@@ -51,13 +51,16 @@ static struct
 	int display_disjuncts;
 	int display_morphology;
 	int display_wordgraph;
-} local, local_saved;
+
+	panic_options panic;
+} local, saved_defaults;
 
 static const char *value_type[] =
 {
 	"(integer) ", "(Boolean) ", "(float) ", "(string) ", "(command) ", ""
 };
 
+static int panic_variables_cmd(const Switch*, int);
 static int variables_cmd(const Switch*, int);
 static int file_cmd(const Switch*, int);
 static int help_cmd(const Switch*, int);
@@ -83,14 +86,20 @@ Switch default_switches[] =
 	{"morphology", Bool, "Display word morphology",         &local.display_morphology},
 	{"null",       Bool, "Allow null links",                &local.allow_null},
 	{"panic",      Bool, "Use of \"panic mode\"",           &local.panic_mode},
+	{"panic_short", Int, "Max length of all links",     &local.panic.short_length},
+	{"panic_cost-max", Float, "Largest cost to be considered", &local.panic.max_cost},
+	{"panic_limit",     Int, "The maximum linkages processed", &local.panic.linkage_limit},
+	{"panic_max-null-count", Int, "Max number of null links allowed", &local.panic.max_null_count},
+	{"panic_spell",     Int, "Up to this many spell-guesses per unknown word", &local.panic.spell_guess},
+	{"panic_timeout",   Int, "Abort panic parsing after this many seconds", &local.panic.timeout},
 	{"postscript", Bool, "Generate postscript output",      &local.display_postscript},
 	{"ps-header",  Bool, "Generate postscript header",      &local.display_ps_header},
 	{"rand",       Bool, "Use repeatable random numbers",   &local.repeatable_rand},
 	{"short",      Int,  "Max length of short links",       &local.short_length},
 #if defined HAVE_HUNSPELL || defined HAVE_ASPELL
 	{"spell",      Int, "Up to this many spell-guesses per unknown word", &local.spell_guess},
-	{"test",       String, "Comma-separated test features", &local.test},
 #endif /* HAVE_HUNSPELL */
+	{"test",       String, "Comma-separated test features", &local.test},
 	{"timeout",    Int,  "Abort parsing after this many seconds", &local.timeout},
 #ifdef USE_SAT_SOLVER
 	{"use-sat",    Bool, "Use Boolean SAT-based parser",    &local.use_sat_solver},
@@ -102,6 +111,7 @@ Switch default_switches[] =
 	{"exit",       Cmd,  "Exit the program",                       exit_cmd},
 	{"file",       Cmd,  "Read input from the specified filename", file_cmd},
 	{"help",       Cmd,  "List the commands and what they do",     help_cmd},
+	{"panic_variables", Cmd,  "List panic mode setup",  panic_variables_cmd},
 	{"quit",       Cmd,  UNDOC "Exit the program",                 exit_cmd},
 	{"variables",  Cmd,  "List user-settable variables and their functions", variables_cmd},
 	{"!",          Cmd,  UNDOC "Print information on dictionary words", info_cmd},
@@ -116,12 +126,12 @@ static void put_opts_in_local_vars(Command_Options *);
 void save_default_opts(Command_Options *copts)
 {
 	put_opts_in_local_vars(copts);
-	local_saved = local;
+	saved_defaults = local;
 }
 
 static void restore_default_local_vars(void)
 {
-	local = local_saved;
+	local = saved_defaults;
 }
 
 /**
@@ -216,7 +226,7 @@ static const char *switch_value_description(const Switch *as)
  * Return a static buffer with a string value of the given switch.
  *
  * Since the static buffer is overwritten on each call, this function
- * should not use more than once as an argument of the same function.
+ * should not be used more than once as an argument of the same function.
  */
 static const char *switch_value_string(const Switch *as)
 {
@@ -258,7 +268,7 @@ static const char *switch_value_string(const Switch *as)
 #define HELPFILE_LANG_TEMPLATE_SIZE (sizeof(HELPFILE_LANG_TEMPLATE)-1)
 #define HELPFILE_TEMPLATE_SIZE \
 	(sizeof(HELPFILE_BASE HELPFILE_EXT)+HELPFILE_LANG_TEMPLATE_SIZE)
-#define D_USER_FILES 3 /* Debug level for files */
+#define D_USER_FILES 4 /* Debug level for files, see error.h. */
 #define DEFAULT_HELP_LANG "en"
 
 static FILE *help_file;
@@ -440,12 +450,12 @@ void display_1line_help(const Switch *sp, bool is_completion)
 
 static void display_help(const Switch *sp, Command_Options *copts)
 {
-	char line[MAX_INPUT]; /* Maximum number of character in a help file line */
+	char line[MAX_INPUT_LINE]; /* Maximum number of character in a help file line */
 
 	display_1line_help(sp, /*is_completion*/false);
 	if (Cmd != sp->param_type)
 	{
-		printf("Current value: %s\n",switch_value_string(sp));
+		printf("Current value: %s\n", switch_value_string(sp));
 		restore_default_local_vars();
 		printf("Default value: %s\n", switch_value_string(sp));
 		put_opts_in_local_vars(copts);
@@ -468,7 +478,7 @@ static void display_help(const Switch *sp, Command_Options *copts)
 			 * in a hope that the same help text can be used for the
 			 * language bindings too (which have different option names
 			 * from historical reasons).
-			 * Note: We suppose the lines are not longer than MAX_INPUT.
+			 * Note: We suppose the lines are not longer than MAX_INPUT_LINE.
 			 * Longer lines may render the help text incorrectly.
 			 * FIXME: Add command reference notation in the help text if
 			 * needed.
@@ -511,7 +521,7 @@ static void display_help(const Switch *sp, Command_Options *copts)
 		while (NULL != fgets(line, sizeof(line), hf))
 		{
 			const size_t len = strlen(line);
-			if ((MAX_INPUT == len-1) && '\n' != line[MAX_INPUT-2])
+			if ((MAX_INPUT_LINE == len-1) && '\n' != line[MAX_INPUT_LINE-2])
 			{
 				prt_error("Warning: Help-file text line too long at offset %ld\n",
 							 ftell(hf));
@@ -569,28 +579,105 @@ static int help_cmd(const Switch *uc, int n)
 	printf("Special commands always begin with \"!\".  Command and variable names\n");
 	printf("can be abbreviated.  Here is a list of the commands:\n\n");
 
-	printf(" !help command   Show a detailed help for the given command.\n");
+	printf(" !help command    Show a detailed help for the given command.\n");
 	for (int i = 0; uc[i].string != NULL; i++)
 	{
 		if (Cmd != uc[i].param_type) continue;
 		if (UNDOC[0] == uc[i].description[0]) continue;
-		printf(" !%-14s ", uc[i].string);
+		printf(" !%-15s ", uc[i].string);
 		printf("%s.\n", uc[i].description);
 	}
 
 	printf("\n");
-	printf(" !!<string>      Print all the dictionary words that match <string>.\n");
-	printf("                 A wildcard * may be used to find multiple matches.\n");
-	printf("                 Issue \"!help !\" for more details.\n");
+	printf(" !!<string>       Print all the dictionary words that match <string>.\n");
+	printf("                  A wildcard * may be used to find multiple matches.\n");
+	printf("                  Issue \"!help !\" for more details.\n");
 	printf("\n");
-	printf(" !<var>          Toggle the specified Boolean variable.\n");
-	printf(" !<var>=<val>    Assign that value to that variable.\n");
+	printf(" !<var>           Toggle the specified Boolean variable.\n");
+	printf(" !<var>=<val>     Assign that value to that variable.\n");
 
 	print_url_info();
 
 	return 'c';
 }
 
+void setup_panic_parse_options(Command_Options *copts, int sentence_length)
+{
+	Parse_Options opts = copts->popts;
+
+	parse_options_reset_resources(opts);
+
+	parse_options_set_max_parse_time(opts, local.panic.timeout);
+
+	/* Use panic_cost if it is greater than the current cost. */
+	parse_options_set_disjunct_cost(opts,
+	   MAX(local.max_cost, local.panic.max_cost));
+
+	/* In case !null=0, min_null_count and max_null_count are already 0 as
+	 * needed. In case !null=1, the parsing is done by default in 2 steps,
+	 * when on the first step nulls are not allowed and in the second step
+	 * nulls are allowed, and any of these steps may fail (the code is
+	 * currently kept this way to make code tests easy). Since the code is
+	 * not structured to potentially make a 2-step panic parsing, the panic
+	 * parsing for the case of !null=1 will be done in one step.
+	 * Note that min_null_count is always already set as needed. */
+
+	if (copts->allow_null)
+	{
+		/* Restrict max_null_count. */
+		parse_options_set_max_null_count(opts,
+	      MIN(sentence_length, local.panic.max_null_count));
+	}
+
+	/* Restrict short_length. */
+	parse_options_set_short_length(opts,
+	   MIN(local.short_length, local.panic.short_length));
+	parse_options_set_all_short_connectors(opts, 1);
+
+	/* Restrict spell. */
+	parse_options_set_spell_guess(opts,
+	   MIN(local.spell_guess, local.panic.spell_guess));
+
+	if (local.verbosity > 1)
+	{
+		int n = 0;
+		fprintf(stdout, "Panic mode setup: %n", &n); /* Remember alignment. */
+		fprintf(stdout,
+		        "!cost-max=%.2f !limit=%d !timeout=%d " "!spell=%d !short=%d\n"
+		        "%*s(all_short=%d min_null_count=%d max_null_count=%d)\n",
+		        parse_options_get_disjunct_cost(opts),
+		        parse_options_get_linkage_limit(opts),
+		        parse_options_get_max_parse_time(opts),
+		        parse_options_get_spell_guess(opts),
+		        parse_options_get_short_length(opts),
+		        n, "",                               /* Align to legend length. */
+		        parse_options_get_all_short_connectors(opts),
+		        parse_options_get_min_null_count(opts),
+		        parse_options_get_max_null_count(opts));
+	}
+}
+
+static int panic_variables_cmd(const Switch *uc, int n)
+{
+	printf(" Variable             Controls                                      Value\n");
+	printf(" --------             --------                                      -----\n");
+	for (int i = 0; uc[i].string != NULL; i++)
+	{
+		if (Cmd == uc[i].param_type) continue;
+		if (strncasecmp("panic_", uc[i].string, 6) != 0) continue;
+		if (UNDOC[0] == uc[i].description[0]) continue;
+		printf(" %-21s", uc[i].string);
+		printf("%-*s", FIELD_WIDTH(uc[i].description, 46), uc[i].description);
+		printf("%5s", switch_value_string(&uc[i]));
+		printf("%s\n", switch_value_description(&uc[i]));
+	}
+
+	printf("\n");
+	printf("Set a variable as so: \"!panic_timeout=10\".\n");
+	printf("Get detailed help on a variable with: \"!help var\".\n");
+	return 'c';
+}
+
 static int variables_cmd(const Switch *uc, int n)
 {
 	printf(" Variable     Controls                                          Value\n");
@@ -598,6 +685,7 @@ static int variables_cmd(const Switch *uc, int n)
 	for (int i = 0; uc[i].string != NULL; i++)
 	{
 		if (Cmd == uc[i].param_type) continue;
+		if (strncasecmp("panic_", uc[i].string, 6) == 0) continue;
 		if (UNDOC[0] == uc[i].description[0]) continue;
 		printf(" %-13s", uc[i].string);
 		printf("%-*s", FIELD_WIDTH(uc[i].description, 46), uc[i].description);
@@ -629,55 +717,76 @@ static int info_cmd(const Switch *uc, int n)
 	return 'c';
 }
 
+const char helpmsg[] = "Type \"!help\" or \"!variables\".";
+
+/**
+ * Handle a request for a particular command help.
+ */
+static int handle_help_command(const Switch *as, char *line,
+                               Command_Options *copts)
+{
+	char *dupline = strdup(line);
+	int count, j;
+	int rc = 0;
+
+	/* If we are here, it is not a command-line parameter. */
+	char *s = strtok(dupline, WHITESPACE);
+	if ((s != NULL) && strncasecmp(s, "help", strlen(s)) == 0)
+	{
+		s = strtok(NULL, WHITESPACE);
+		if (s != NULL)
+		{
+			/* This is a help request for the command name at s. */
+			j = -1; /* command index */
+			count = 0;  /* number of matching commands */
+
+			/* Is it a unique abbreviation? */
+			for (int i = 0; as[i].string != NULL; i++)
+			{
+				if (strncasecmp(s, as[i].string, strlen(s)) == 0)
+				{
+					j = i;
+					if (strlen(as[i].string) == strlen(s))
+					{
+						count = 1;
+						break;
+					}
+					count++;
+				}
+			}
+
+			if (count == 1)
+			{
+				display_help(&as[j], copts);
+				rc = 'c';    /* Command done. */
+			}
+			else
+			{
+				rc = -1;     /* Error indication. */
+				if (count > 1)
+					prt_error("Ambiguous command: \"%s\".  %s\n", s, helpmsg);
+				else
+					prt_error("Undefined command: \"%s\".  %s\n", s, helpmsg);
+			}
+		}
+	}
+
+	free(dupline);
+	return rc;
+}
+
 static int x_issue_special_command(char * line, Command_Options *copts, Dictionary dict)
 {
 	char *s, *x, *y;
 	int count, j;
 	const Switch *as = default_switches;
-	const char helpmsg[] = "Type \"!help\" or \"!variables\".";
 
-	/* Handle a request for a particular command help. */
-	if (NULL != dict)
+	line = &line[strspn(line, WHITESPACE)]; /* Trim initial whitespace */
+
+	if (NULL != dict) /* No dict if we are called from the command line */
 	{
-		char *dupline = strdup(line);
-		/* If we are here, it is not a command-line parameter. */
-		s = strtok(dupline, WHITESPACE);
-		if ((s != NULL) && strncasecmp(s, "help", strlen(s)) == 0)
-		{
-			s = strtok(NULL, WHITESPACE);
-			if (s != NULL)
-			{
-				/* This is a help request for the command name at s. */
-				j = -1; /* command index */
-				count = 0;  /* number of matching commands */
-
-				/* Is it a unique abbreviation? */
-				for (int i = 0; as[i].string != NULL; i++)
-				{
-					if (strncasecmp(s, as[i].string, strlen(s)) == 0)
-					{
-						count++;
-						j = i;
-					}
-				}
-
-				if (count == 1)
-				{
-					free(dupline);
-					display_help(&as[j], copts);
-					return 'c';
-				}
-
-				if (count > 1)
-					prt_error("Ambiguous command: \"%s\".  %s\n", s, helpmsg);
-				else
-					prt_error("Undefined command: \"%s\".  %s\n", s, helpmsg);
-
-				free(dupline);
-				return -1;
-			}
-		}
-		free(dupline);
+		int rc = handle_help_command(as, line, copts);
+		if (rc != 0) return rc;
 	}
 
 	s = line;
@@ -710,44 +819,59 @@ static int x_issue_special_command(char * line, Command_Options *copts, Dictiona
 		return 'c';
 	}
 
+	if (strchr(s, '=') == NULL)
+	{
+		j = -1;
+		count = 0;
+
+		/* Look for Boolean flippers or command abbreviations. */
+		for (int i = 0; as[i].string != NULL; i++)
+		{
+			if (((Bool == as[i].param_type) || (Cmd == as[i].param_type)) &&
+			    strncasecmp(s, as[i].string, strcspn(s, WHITESPACE)) == 0)
+			{
+				j = i;
+				if (strlen(as[i].string) == strcspn(s, WHITESPACE))
+				{
+					count = 1;
+					break;
+				}
+				if (UNDOC[0] == as[i].description[0]) continue;
+				count++;
+			}
+		}
+
+		if (count > 1)
+		{
+			prt_error("Ambiguous command \"%s\".  %s\n", s, helpmsg);
+			return -1;
+		}
+		if (count == 1)
+		{
+			/* Flip Boolean value. */
+			if (Bool == as[j].param_type)
+			{
+				size_t junk = strcspn(s, WHITESPACE);
+				if (junk != strlen(s))
+				{
+					prt_error("Junk after a boolean variable: \"%s\".  %s\n",
+					          &s[junk], helpmsg);
+					return -1;
+				}
+				setival(as[j], (0 == ival(as[j])));
+				int undoc = !!(UNDOC[0] == as[j].description[0]);
+				printf("%s turned %s.\n",
+				       as[j].description+undoc, (ival(as[j]))? "on" : "off");
+				return 'c';
+			}
+
+			/* Found an abbreviated, but it wasn't a Boolean.
+			 * It means it is a user command, to be handled below. */
+			return ((int (*)(const Switch*, int)) (as[j].ptr))(as, j);
+		}
+	}
+
 	clean_up_string(line);
-	j = -1;
-	count = 0;
-
-	/* Look for Boolean flippers or command abbreviations. */
-	for (int i = 0; as[i].string != NULL; i++)
-	{
-		if (((Bool == as[i].param_type) || (Cmd == as[i].param_type)) &&
-		    strncasecmp(s, as[i].string, strlen(s)) == 0)
-		{
-			if ((UNDOC[0] == as[i].description[0]) &&
-			    (strlen(as[i].string) != strlen(s))) continue;
-			count++;
-			j = i;
-		}
-	}
-
-	if (count > 1)
-	{
-		prt_error("Ambiguous command \"%s\".  %s\n", s, helpmsg);
-		return -1;
-	}
-	else if (count == 1)
-	{
-		/* Flip Boolean value. */
-		if (Bool == as[j].param_type)
-		{
-			setival(as[j], (0 == ival(as[j])));
-			int undoc = !!(UNDOC[0] == as[j].description[0]);
-			printf("%s turned %s.\n",
-			       as[j].description+undoc, (ival(as[j]))? "on" : "off");
-			return 'c';
-		}
-
-		/* Found an abbreviated, but it wasn't a Boolean.
-		 * It means it is a user command, to be handled below. */
-		return ((int (*)(const Switch*, int)) (as[j].ptr))(as, j);
-	}
 
 	/* Test here for an equation i.e. does the command line hold an equals sign? */
 	for (x=s; (*x != '=') && (*x != '\0'); x++)
@@ -761,14 +885,19 @@ static int x_issue_special_command(char * line, Command_Options *copts, Dictiona
 
 		/* Figure out which command it is .. it'll be the j'th one */
 		j = -1;
+		count = 0;
 		for (int i = 0; as[i].string != NULL; i++)
 		{
 			if (Cmd == as[i].param_type) continue;
 			if (strncasecmp(x, as[i].string, strlen(x)) == 0)
 			{
-				if ((UNDOC[0] == as[i].description[0]) &&
-				    (strlen(as[i].string) != strlen(x))) continue;
 				j = i;
+				if (strlen(as[i].string) == strlen(x))
+				{
+					count = 1;
+					break;
+				}
+				if (UNDOC[0] == as[i].description[0]) continue;
 				count ++;
 			}
 		}
@@ -796,7 +925,7 @@ static int x_issue_special_command(char * line, Command_Options *copts, Dictiona
 			if ((val < 0) ||
 			    ((as[j].param_type == Bool) && (val != 0) && (val != 1)))
 			{
-				prt_error("Error: Invalid value %s for variable \"%s\". %s\n",
+				prt_error("Error: Invalid value \"%s\" for variable \"%s\". %s\n",
 				          y, as[j].string, helpmsg);
 				return -1;
 			}
@@ -812,7 +941,7 @@ static int x_issue_special_command(char * line, Command_Options *copts, Dictiona
 			double val = strtod(y, &err);
 			if (('\0' == *y) || ('\0' != *err))
 			{
-				prt_error("Error: Invalid value %s for variable \"%s\". %s\n",
+				prt_error("Error: Invalid value \"%s\" for variable \"%s\". %s\n",
 				          y, as[j].string, helpmsg);
 				return -1;
 			}
@@ -897,9 +1026,11 @@ static void put_opts_in_local_vars(Command_Options* copts)
 	local.display_links = copts->display_links;
 
 	local.display_morphology = parse_options_get_display_morphology(opts);
+
+	local.panic = copts->panic;
 }
 
-static void put_local_vars_in_opts(Command_Options* copts)
+void put_local_vars_in_opts(Command_Options* copts)
 {
 	Parse_Options opts = copts->popts;
 	parse_options_set_verbosity(opts, local.verbosity);
@@ -935,19 +1066,13 @@ static void put_local_vars_in_opts(Command_Options* copts)
 	copts->display_bad = local.display_bad;
 	copts->display_disjuncts = local.display_disjuncts;
 	copts->display_links = local.display_links;
+
+	copts->panic = local.panic;
 }
 
 int issue_special_command(const char * line, Command_Options* opts, Dictionary dict)
 {
 	int rc;
-	Parse_Options save = NULL;
-
-	if (strncmp(line, "panic_", 6) == 0)
-	{
-		line += 6;
-		save = opts->popts;
-		opts->popts = opts->panic_opts;
-	}
 
 	put_opts_in_local_vars(opts);
 	char *cline = strdup(line);
@@ -959,7 +1084,6 @@ int issue_special_command(const char * line, Command_Options* opts, Dictionary d
 	put_opts_in_local_vars(opts);
 	free(cline);
 
-	if (save) opts->popts = save;
 	return rc;
 }
 
@@ -968,7 +1092,6 @@ Command_Options* command_options_create(void)
 {
 	Command_Options* co = malloc(sizeof (Command_Options));
 	co->popts = parse_options_create();
-	co->panic_opts = parse_options_create();
 
 	/* "Unlimited" screen width when writing to a file, auto-updated
 	 * later, when writing to a tty. */
@@ -976,7 +1099,7 @@ Command_Options* command_options_create(void)
 	co->allow_null = true;
 	co->batch_mode = false;
 	co->echo_on = false;
-	co->panic_mode = false;
+	co->panic_mode = true;
 	co->display_on = true;
 	co->display_walls = false;
 	co->display_postscript = false;
@@ -987,12 +1110,18 @@ Command_Options* command_options_create(void)
 	co->display_disjuncts = false;
 	co->display_links = false;
 	co->display_wordgraph = 0;
+
+	co->panic.max_cost = 4.0f;
+	co->panic.linkage_limit = 1000;
+	co->panic.max_null_count = 10;
+	co->panic.short_length = 12;
+	co->panic.spell_guess = 0;
+	co->panic.timeout = 30;
 	return co;
 }
 
 void command_options_delete(Command_Options* co)
 {
-	parse_options_delete(co->panic_opts);
 	parse_options_delete(co->popts);
 	free(co);
 }
